import asyncio
import logging

import aiohttp

from postbin.v2 import errors
from postbin.v2.errors import FailedTest, HTTPException

__version__ = "2.0.2a"

_HEADERS = {"User-Agent": f"PostBin (https://github.com/dragdev-studios/postbin)/{__version__})"}
_FALLBACKS = [
    "https://haste.clicksminuteper.net",
    "https://hastebin.com",
    "https://mystb.in",
    "https://paste.pythondiscord.com",
    "https://haste.unbelievaboat.com",
    "https://hst.sh",
    "https://hasteb.in"
]


class ConfigOptions:
    """Class similar to **kwargs except takes up less room."""
    def __init__(self, **kwargs):
        self.test_urls_first = kwargs.pop("test_urls_first", False)
        self.return_full_url = kwargs.pop("return_full_url", True)
        self.ignore_http_errors = kwargs.pop("ignore_http_errors", False)


class AsyncHaste:
    """
    Haste operations, centralised.

    All methods of this class, that need to be, are async.
    """
    def __init__(self, t: str = None, session: aiohttp.ClientSession = None):
        """Creates the class. You shouldn't provide arguments (other than [t]ext)"""
        self.text = t
        self.session = session

    async def find_working_fallback(self, retries_per_url: int = 3):
        for url in _FALLBACKS:
            try:
                await self._head(url)
            except aiohttp.ClientResponseError:
                continue
            else:
                return url
        raise ConnectionError("Unable to connect anywhere. Are you sure you're online?")

    async def _get_session(self) -> aiohttp.ClientSession:
        if not self.session or self.session.closed:
            self.session = aiohttp.ClientSession()
        return self.session

    async def _head(self, url, retries=3):
        last_response = None
        for i in range(retries+1):
            try:
                async with await self._get_session() as session:
                    async with session.head(url) as response:
                        # some services may not support HEAD requests,so we can just GET if not.
                        # We never download the content anyway, its more saving the server's bandwidth.
                        # Because we're not assholes.
                        if response.status == 405:
                            async with session.get(url) as embedded_response:
                                last_response = embedded_response
                                response.raise_for_status()
                                return True
                        else:
                            last_response = response
                            response.raise_for_status()
                            return True
            except (aiohttp.ClientError, ConnectionError):
                continue
        return last_response

    def __del__(self):
        self.close_session()

    def close_session(self):
        """
        Simple method that performs cleanup of the class.

        This is called automatically by garbage collection (__del__)

        :return: Optional[asyncio.Task] - if task is provided, it is closing the aiohttp session.
        """
        if not hasattr(self, "session") or not self.session or self.session.closed:
            return
        loop = asyncio.get_event_loop()
        if loop.is_running():
            return loop.create_task(self.session.close(), name="PostBin cleanup task")
        loop.run_until_complete(self.session.close())  # this could cause issues with other async apps
        # if the edge-case where it tries to do the same as us
        # but we got the loop first
        return

    async def _post(self, url, text, **kwargs):
        try:
            async with await self._get_session() as session:
                async with session.post(url,data=text, **kwargs) as response:
                    retry_after = response.headers.get("retry_after") or response.headers.get("x-retry-after")
                    if response.status == 429 and retry_after:
                        await asyncio.sleep(float(retry_after))
                        return await self._post(url, text, **kwargs)
                    if response.status == 400:
                        data = await response.json()
                        if data["message"].lower() == "document exceeds maximum length.":
                            raise errors.TextTooLarge(response, message=data["message"] + "\nText: " + text)
                    elif response.status == 413:
                        raise errors.TextTooLarge(response, message="Text: " + text)
                    if response.status not in [200, 201]:  # removed 202: That is processing, not complete.
                        raise HTTPException(response)
                    return (await response.json())["key"]
        except (aiohttp.ServerDisconnectedError, aiohttp.ClientConnectorError, aiohttp.ClientOSError) as e:
            raise errors.OfflineServer(None, message="Exception while connecting - assuming dead host.") from e

    async def post(self, text: str = None, config: ConfigOptions = ConfigOptions(), *, timeout: float = 30.0,
                   retries: int = 3, url: str = "auto"):
        """
        Creates a haste URL, returning the URL of the new haste.

        :param text: The text to post. Defaults to the text provided/set in self.__init__.
        :param config: The configuration.
        :param timeout: After how long to stop trying to post to a certain URL, and find a fallback.
        :param retries: How many times to attempt to post to the current URL before finding a fallback.
        :param url: The BASE url to post to. If "auto" (default), this will try each url until it works.
        :return: the completed URL or just the key if specified in Config.
        """
        if url != "auto" and config.test_urls_first:
            response = await self._head(url, retries)
            if not response:
                raise FailedTest(response)
        elif url == "auto":
            url = await self.find_working_fallback(retries)
        try:
            res = await asyncio.wait_for(self._post(url + "/documents", text or self.text, headers=_HEADERS),
                                         timeout=timeout)
        except Exception as e:
            if config.ignore_http_errors:
                return ""
            raise e
        return res if not config.return_full_url else url + "/" + res

    async def raw(self, key: str, *, url: str = "auto", timeout: float = 30.0, retries_per_url: int = 3,
                  encoding: str = "utf-8"):
        """
        Gets the raw text of a haste.

        Note that the parameters are different from post, in that they apply per URL fetched (if url is auto)

        :param key: the key (after .tld/, e.g hastebin.com/{key})
        :param url: the URL to find the haste from. if "auto" (default), will search all fallbacks for it.
        :param timeout: The timeout to request a document from the servers. If this is hit, instead of raising timeout error, just skips to the next one.
        :param retries_per_url: how may times to retry a URL (unless it returned 404). set to 0 to disable.
        :param encoding: The encoding to encode the text in. Defaults to utf-8. | Version added: 2.0.2a
        :return: the found text, __or None if not found__.
        """
        async def get(URL):
            async with session.get(URL + "/raw/" + key) as response:
                if response.status == 404:
                    return None
                if response.status == 200:
                    return await response.text(encoding=encoding or "utf-8", errors="replace")

        session = await self._get_session()
        if url.lower() == "auto":
            for _URL in _FALLBACKS:
                res = await get(_URL)
                if res:
                    return res
            return None
        return await get(url + "/raw/" + key)


async def postAsync(text: str, *, url: str = "auto", config: ConfigOptions = ConfigOptions(), timeout: float = 30.0,
                    retries: int = 3):
    """Alias function for AsyncHaste().post(...)"""
    return await AsyncHaste().post(text, url=url, config=config, timeout=timeout, retries=retries)


def postSync(text: str, *, url: str = "auto", config: ConfigOptions = ConfigOptions(), timeout: float = 30.0,
             retries: int = 3):
    """
    Alias function for SyncHaste().post(...).

    WARNING! This relies on the asyncio event loop NOT being in use. If it is, this returns an asyncio.Task
    """
    paster = AsyncHaste(text)
    f = paster.post(None, config, timeout=timeout, retries=retries, url=url)
    loop = asyncio.get_event_loop()
    if loop.is_running():
        return loop.create_task(f, name="Paste-" + str(id(paster)))
    return loop.run_until_complete(f)



from typing import Union
try:
    from requests import Response
except ImportError:
    Response = None

try:
    from aiohttp import ClientResponse
except ImportError:
    ClientResponse = None
#

class HTTPException(Exception):
    """
    Represents a HTTP Exception.

    Attributes:
        response - Union[ClientResponse, Response]: the response that was returned.
        message - Optional[str]: A custom message as to why the error was raised.
        status - int: The resolved status (so you don't have to dig between the two classes)
    """
    def __init__(self, response: Union[Response, None, ClientResponse], *args, message: str = None, **kwargs):
        """
        Creates the HTTPException

        :param response: the Response object
        :param args: args to pass to super
        :param message: a custom message as to why the error was raised
        :param kwargs: doesn't seem like much to me.
        """
        super().__init__(kwargs, *args)
        self.response = response
        self.message = message
        try:
            self.status = response.status if isinstance(response, ClientResponse) else response.status_code
        except AttributeError:
            self.status = -1  # response was None.

    def __str__(self):
        if self.message:
            return self.__class__.__name__ + ": " + self.message  # for dynamic stuff
        return "{0.__class__.__name__}: Got status {0.status} on URL {0.response.url}.".format(self)

    def __repr__(self):
        return f"{self.__class__.__name__}(response={self.response} message={self.message} status={self.status})"


class FailedTest(HTTPException):
    """
    Raised when a URL test fails.
    """


class OfflineServer(HTTPException):
    """The server provided wasn't online or valid."""


class TextTooLarge(HTTPException):
    """Raised when the text provided to post was too large.

    Bear in mind most template haste services have a paste size limit of 1mb."""



## Note to editors and future/current contributors:  ##
# this file is very messy.                            #
# Any contributions to tidying it up would be greatly #
# appriciated.                                        #
#                                                     #
# - eek                                               #
# # # # # # # # # # # # # # # # # # # # # # # # # # # #

import typing

try:
    import requests
    RR = requests.Response
except ImportError:
    requests = None
    RR = None

try:
    import aiohttp
    CR = aiohttp.ClientResponse
except ImportError:
    aiohttp = None
    CR = None
import time
import asyncio

_FALLBACKS = [
    "https://mystb.in",
    "https://haste.unbelievaboat.com",
    "https://paste.pythondiscord.com",    
    "https://hastebin.com"
]
#_HASTE_URLS_FOR_REGEX = '|'.join(_FALLBACKS[8:]).replace(".", "\\.")
# _HASTE_URLS_RAW = "(https://|http://)?({})/(raw/)?(?P<key>.+)".format(_HASTE_URLS_FOR_REGEX)
# _HASTE_REGEX = re.compile(_HASTE_URLS_RAW)

# we don't need the above anymore - Will be removed in the future.


class ResponseError(Exception):
    """Generic class raised when contacting the server failed."""
    def __init__(self, response: typing.Union[RR,CR]):
        self.raw_response = response
        if isinstance(response, requests.Response):
            self.status = response.status_code
        else:
            self.status = response.status

class NoFallbacks(Exception):
    """Raised when no fallback could be contacted."""

class NoMoreRetries(NoFallbacks):
    """Raised when we ran out of attempts to post."""


def findFallBackSync(verbose: bool = False):
    """Tries to find a fallback URL, if haste.clicksminuteper.net isn't working."""
    if not requests:
        raise RuntimeError("You need to install requests to be able to use findFallBackSync.")
    with requests.Session() as session:
        n = 0
        for n, url in enumerate(_FALLBACKS, 1):
            if verbose:
                print(f"Trying service {n}/{len(_FALLBACKS)}",end="\r")
            try:
                response = session.post(url+"/documents", data="")
            except:
                if verbose:
                    print(f"Service {n}/{len(_FALLBACKS)} failed. trying again.", end="\r")
                continue
            if response.status_code != 200:
                continue
            else:
                if verbose:
                    print(f"{url} ({n}) worked. Using that.")
                break
        else:
            if verbose:
                print("No functional URLs could be found. Are you sure you're online?")
            raise NoFallbacks()
        return url

async def findFallBackAsync(verbose: bool = False):
    """Same as findFallBackSync, but just async."""
    if not aiohttp:
        raise RuntimeError("You need to install aiohttp to be able to use findFallBackAsync.")
    async with aiohttp.ClientSession() as session:
        for n, url in enumerate(_FALLBACKS, 1):
            if verbose:
                print(f"Trying service {n}/{len(_FALLBACKS)} (URL {url})",end="\r")
            try:
                async with session.post(url+"/documents", data="") as response:
                    if response.status != 200:
                        continue
                    else:
                        if verbose:
                            print(f"{url} ({n}) worked. Using that.")
                        return url
            except:
                if verbose:
                    print(f"Service {n}/{len(_FALLBACKS)} failed. trying again.", end="\r")
                continue
        else:
            if verbose:
                print("No functional URLs could be found. Are you sure you're online?")
            raise NoFallbacks()


def post(sync: bool = False, *, content: str, url: str = None, retry: int = 5, find_fallback: bool = True):
    """
    Creates a new haste.

    If :param:`sync` is ``False``, this function returns a coroutine that you need to await.

    Examples: ::

        # You can run this in ipython or `python3 -m asyncio`.
        from postbin import post
        sync_url = post(True, content="Hello")

        async def main():
            async_url = await post(False, content="Hello")
        await main()
    """
    if sync:
        return postSync(content, url=url, retry=retry, find_fallback_on_unavailable=find_fallback)
    else:
        return postAsync(content, url=url, retry=retry, find_fallback_on_unavailable=find_fallback)


# noinspection PyIncorrectDocstring
def postSync(content:  str, *, url: str = None, retry: int = 5, find_fallback_on_unavailable: bool = True,
             find_fallback_on_retry_runout: bool = False):
    """
    Creates a new haste

    :param content: Union[str, Iterable] - the content to post to hastebin. If this is a list, it will join with a newline.
    :keyword url: the custom URL to post to. Defaults to CMP Haste.
    :keyword retry: the number of times to retry. Pass `0` to disable
    :keyword find_fallback_on_unavailable: Whether or not to find a fallback or give up if the url fails to return.
    :keyword find_fallback_on_retry_runout: if True, instead of raising NoMoreRetries(), find a fallback instead.
    :raise RuntimeError: requests is not installed or you provided a retry value above 1000 (which would raise recursion error)
    :raise TypeError: Either the provided `content` was not string/iterable, or you disabled find_..._unavailable.
    :return: the returned URL
    """
    if not requests:
        raise RuntimeError("requests must be installed if you want to be able to run postSync.")
    if not isinstance(content, str):
        content = repr(content)
    url = url or _FALLBACKS[0]
    with requests.Session() as session:
        try:
            response = session.post(url+"/documents", data=content)
            if response.status_code == 503:
                return postSync(content, url=findFallBackSync(True), find_fallback_on_retry_runout=True)
            if response.status_code != 200:
                raise ResponseError(response)
            if response.headers.get("Content-Type", "").lower() != "application/json":
                return postSync(content, url=findFallBackSync(True), find_fallback_on_retry_runout=True)
            key = response.json()["key"]
        except (requests.ConnectionError, ConnectionError):
            print(url, "is unable. Finding a fallback...")
            return postSync(content, url=findFallBackSync(True), find_fallback_on_retry_runout=True)
        except Exception as e:
            if retry <= 0:
                if find_fallback_on_retry_runout:
                    print(url, "is unavailable. Finding a fallback...")
                    return postSync(content, url=findFallBackSync(True), find_fallback_on_retry_runout=True)
                raise NoMoreRetries()
            retry -= 1
            if find_fallback_on_unavailable:
                return postSync(content, url=url, retry=retry, find_fallback_on_unavailable=True,
                                find_fallback_on_retry_runout=True)
            raise TypeError("Unable to create a haste with the provided URL.")
    return url+"/"+key

async def postAsync(content: str, *, url: str = None, retry: int = 5, find_fallback_on_unavailable: bool = True,
                    find_fallback_on_retry_runout: bool = False):
    """The same as :func:postSync, but async."""
    if not aiohttp:
        raise RuntimeError("aiohttp must be installed if you want to be able to run postAsync.")
    if not isinstance(content, str):
        content = repr(content)
    url = url or _FALLBACKS[0]
    async with aiohttp.ClientSession() as session:
        try:
            async with session.post(url+"/documents", data=content) as response:
                if response.status == 503:
                    return await postAsync(content, url=await findFallBackAsync(True), find_fallback_on_retry_runout=True)
                if response.status != 200:
                    raise ResponseError(response)
                if response.headers.get("Content-Type", "").lower() != "application/json":
                    return await postAsync(content, url=await findFallBackAsync(True), find_fallback_on_retry_runout=True)
                key = (await response.json())["key"]
                return f"{url}/{key}"
        except aiohttp.ClientConnectionError:
            if find_fallback_on_unavailable:
                print(url, "is unavailable. Finding a fallback...")
                return await postAsync(content, url=await findFallBackAsync(True), find_fallback_on_retry_runout=True)
            raise TypeError("Unable to create a haste with the provided URL")
        except Exception as e:
            print(e)
            if retry <= 0:
                if find_fallback_on_retry_runout:
                    return await postAsync(content, url=await findFallBackAsync(True), find_fallback_on_retry_runout=True)
                raise NoMoreRetries()
            retry -= 1
            # return await postAsync(content, url=url, retry=retry, find_fallback_on_retry_runout=True)
            if find_fallback_on_unavailable:
                return await postAsync(content, url=url, retry=retry, find_fallback_on_unavailable=True,
                                find_fallback_on_retry_runout=True)
            raise TypeError("Unable to create a haste with the provided URL.")
    return url+"/"+key



from setuptools import setup

try:
    with open("README.md") as rfile:
        long_description = rfile.read()
except:
    long_description = "https://github.com/ytcalifax/PyHasteBin (description failed to load)"

setup(
    name='PyHasteBin',
    version='1.2.1',
    packages=['postbin', "postbin.v2"],
    url='https://github.com/ytcalifax/PyHasteBin',
    python_requires=">=3.6",
    license='MIT',
    author='EEKIM10',
    author_email='eek@clicksminuteper.net',
    description='A simple two-function program that POSTs to hastebin',
    long_description=open("README.md").read(),
    long_description_content_type='text/markdown',
    extras_require={"cli": "webbrowser"}
)
#



from os import path
from sys import argv

from postbin import postSync

if len(argv) >= 2:
    text = ' '.join(argv[1:])
else:
    text = input(
        "What text would you like to put on hastebin?\nIf you provide a filepath, it will post that file's content\n\n> ")
if path.exists(text):
    try:
        with open(text) as rfile:
            text = rfile.read()
    except:
        print("Attempted to open a file, but was unable to. Assuming you didn't want to.")
url = postSync(text, retry=10)
try:
    import webbrowser

    webbrowser.open(url, 0)
except ImportError:
    webbrowser = None
    print("Tried to open in browser, but webbrowser was not installed.")
    print()
exit(0)
#
